from functools import partial
import inspect
import pandas
from typing import Any, Callable, Dict, Iterable, List, Optional, Type, Union
import warnings


class DynamicSeriesAccessor:
    """
    Pandas Series Accessor for dynamically accessing methods and properties of
    a target class.

    The accessor allows you to call the methods and properties of the specified
    target class on a Pandas Series. Each method or property accessed through the
    accessor returns a new Series, where each element is the result of applying
    the method or property on the target class instance, initialised with the
    value from the original series.

    Optionally, you can Specify valid types for the values of the series; and
    define custom methods that can be accessed through the Series accessor, in
    addition to the methods and properties of the target class.

    Parameters:
    - series (pandas.Series): The Pandas Series to extend.
    - target_class (Type): The class whose methods and properties will be mapped
      to the accessor.
    - excluded_attrs (Optional[Iterable[str]]): An iterable object holding blocked
      attributes of the target class.
    - valid_types (Optional[Iterable[Type]]): Allowed value types in the Pandas
      Series.
    - custom_methods (Optional[Dict[str, Callable]]): Custom methods that can be
      accessed on the Pandas Series.
    """

    def __init__(
        self,
        series: pandas.Series,
        target_class: Type,
        excluded_attrs: Optional[Iterable[str]] = None,
        valid_types: Optional[Iterable[Type]] = None,
        custom_methods: Optional[Dict[str, Callable]] = None,
    ) -> None:
        self._s = series
        self._cls = target_class
        self._excl = excluded_attrs if excluded_attrs else ()
        self._valid_types = valid_types if valid_types else None
        self._custom_methods = custom_methods if custom_methods else {}
        self._validate()

    def _validate(self):
        """Ensure the Series is compatible with the target class.

        Raises an ``AttributeError`` if the Series is empty, contains invalid types,
        or if the target class requires more than one mandatory parameter to initialise.
        """
        if len(self._get_cls_init_params(mandatory_only=True)) > 1:
            raise AttributeError(
                f"{self._cls.__name__!r} requires more than one mandatory initialisation parameter."
            )

        if self._s.empty:
            raise AttributeError(f"{self} only works with non-empty series.")

        if self._valid_types and not self._s.apply(type).isin(self._valid_types).all():
            raise AttributeError(
                f"{self} only works with series of type(s) "
                f"{", ".join(t.__name__ for t in self._valid_types)}."
            )

        if not self._s.apply(self._can_initialise).all():
            raise AttributeError(
                f"{self} only works with series containing values "
                f"that can initialise {self._cls.__name__!r}."
            )

    def _can_initialise(self, value: Any) -> bool:
        """Check if the value can initialise the target class."""
        try:
            self._cls(value)
            return True
        except Exception:
            return False

    def _get_cls_init_params(self, mandatory_only: bool = True) -> Dict[str, Type]:
        """
        Get the init parameters of the target class.

        Parameters:
        - mandatory_only (bool): If True, returns only mandatory parameters (no
          default value). If False, returns all parameters including optional ones.

        Returns:
        - Dict[str, Type]: A dictionary of parameter names and their types.
        """
        init_signature = inspect.signature(self._cls.__init__)
        return {
            param.name: param.annotation
            for param in init_signature.parameters.values()
            if param.name != "self"
            and (not mandatory_only or param.default is param.empty)
        }

    def _wrap_cls_attribute(
        self, name: str
    ) -> Union[pandas.Series, Callable[..., pandas.Series]]:
        """Wraps target class method or property for dynamic invocation on the Series."""
        attr = getattr(self._cls, name)

        # Handle properties
        if not callable(attr):
            return self._s.apply(lambda x: attr.__get__(self._cls(x)))

        # Handle methods
        def method_wrapper(*args, **kwargs) -> pandas.Series:
            return self._s.apply(lambda x: attr(self._cls(x), *args, **kwargs))

        return method_wrapper

    def _wrap_custom_method(self, name: str) -> Callable[..., pandas.Series]:
        """Wraps custom method for dynamic invocation on the Series."""
        custom_method = self._custom_methods[name]

        def custom_method_wrapper(*args, **kwargs) -> pandas.Series:
            return self._s.apply(lambda x: custom_method(x, *args, **kwargs))

        return custom_method_wrapper

    def __dir__(self) -> List[str]:
        """Return a list of available target class attributes and custom methods."""
        class_attrs = dir(self._cls)
        attrs = [attr for attr in class_attrs if attr not in self._excl]
        return attrs + list(self._custom_methods.keys())

    def __getattr__(
        self, name: str
    ) -> Union[Callable[..., pandas.Series], pandas.Series]:
        """Dynamically resolve target class attributes and custom methods."""
        if name not in dir(self):
            raise AttributeError(f"{self!r} does not provide attribute {name!r}.")

        if name in self._custom_methods:
            return self._wrap_custom_method(name)

        return self._wrap_cls_attribute(name)

    def __str__(self) -> str:
        """Return a string representation of the dynamic series accessor."""
        return f"{self._cls.__name__!r} Series Accessor"


def create_dynamic_series_accessor(
    name: str,
    target_class: Type,
    excluded_attrs: Iterable[str] = None,
    valid_types: Optional[Iterable[Type]] = None,
    custom_methods: Optional[Dict[str, Callable]] = None,
) -> None:
    """
    Configures and registers an accessor for dynamic access to methods and
    properties of a target class on Pandas Series.

    This function preconfigures the accessor with the specified target class,
    excluded attributes, valid types, and custom methods. The preconfigured
    accessor is then registered under the given ``name`` as an extension to
    the Pandas Series.

    Parameters:
    - name (str): The name used to register the accessor to the Pandas Series.

    See `DynamicSeriesAccessor` for full documentation.
    """
    accessor = partial(
        DynamicSeriesAccessor,
        target_class=target_class,
        excluded_attrs=excluded_attrs,
        valid_types=valid_types,
        custom_methods=custom_methods,
    )

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("ignore")
        pandas.api.extensions.register_series_accessor(name)(accessor)

        for warning in w:
            if "overriding a preexisting attribute with the same name" not in str(
                warning.message
            ):
                warnings.warn(warning.message, category=warning.category)


def _update_accessor_doc_with_class_details(
    fn_tail_lines_to_exclude: int, cls_head_lines_to_exclude: int
) -> None:
    """
    Updates the documentation of the `create_dynamic_series_accessor` function
    by inserting lines from the `DynamicSeriesAccessor` class's docstring.

    Parameters:
    - fn_tail_lines_to_exclude (int): Number of lines to exclude from the end
      of the function's docstring.
    - cls_head_lines_to_exclude (int): Number of lines to exclude from the
      beginning of the class's docstring.
    """
    fn_doc_lines = create_dynamic_series_accessor.__doc__.strip().splitlines()
    cls_doc_lines = DynamicSeriesAccessor.__doc__.strip().splitlines()

    updated_doc = "\n".join(
        fn_doc_lines[:-fn_tail_lines_to_exclude]
        + cls_doc_lines[cls_head_lines_to_exclude:]
        + fn_doc_lines[-fn_tail_lines_to_exclude:]
    )

    create_dynamic_series_accessor.__doc__ = updated_doc


_update_accessor_doc_with_class_details(2, 15)
